'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

let _stringifySqlAST = (() => {
  var _ref2 = _asyncToGenerator(function* (parent, node, prefix, context, selections, joins, wheres, batchScope) {
    switch (node.type) {
      case 'table':
        if (node.where && (!node.sqlBatch || !parent)) {
          const whereCondition = yield node.where(`${quote(node.as)}`, node.args || {}, context, (0, _shared.quotePrefix)(prefix, '`'));
          if (whereCondition) {
            wheres.push(`${whereCondition}`);
          }
        }

        if (node.sqlJoin) {
          const joinCondition = yield node.sqlJoin(`${quote(parent.as)}`, `${quote(node.as)}`, node.args || {}, context);

          joins.push(`LEFT JOIN ${node.name} AS ${quote(node.as)} ON ${joinCondition}`);
        } else if (node.junctionTable && node.junctionBatch) {
          if (parent) {
            selections.push(`${quote(parent.as)}.${quote(node.junctionBatch.parentKey.name)} AS ${quote((0, _shared.joinPrefix)(prefix) + node.junctionBatch.parentKey.as)}`);
          } else {
            const joinCondition = yield node.junctionBatch.sqlJoin(`${quote(node.junctionTableAs)}`, `${quote(node.as)}`, node.args || {}, context);
            joins.push(`FROM ${node.junctionTable} AS ${quote(node.junctionTableAs)}`, `LEFT JOIN ${node.name} AS ${quote(node.as)} ON ${joinCondition}`);
            wheres.push(`${quote(node.junctionTableAs)}.${quote(node.junctionBatch.thisKey.name)} IN (${batchScope.join(',')})`);
          }
        } else if (node.junctionTable) {
          (0, _assert2.default)(node.sqlJoins, 'Must set "sqlJoins" for a join table.');
          const joinCondition1 = yield node.sqlJoins[0](`${quote(parent.as)}`, `${quote(node.junctionTableAs)}`, node.args || {}, context);
          const joinCondition2 = yield node.sqlJoins[1](`${quote(node.junctionTableAs)}`, `${quote(node.as)}`, node.args || {}, context);

          joins.push(`LEFT JOIN ${node.junctionTable} AS ${quote(node.junctionTableAs)} ON ${joinCondition1}`, `LEFT JOIN ${node.name} AS ${quote(node.as)} ON ${joinCondition2}`);
        } else if (node.sqlBatch) {
          if (parent) {
            selections.push(`${quote(parent.as)}.${quote(node.sqlBatch.parentKey.name)} AS ${quote((0, _shared.joinPrefix)(prefix) + node.sqlBatch.parentKey.as)}`);
          } else {
            joins.push(`FROM ${node.name} AS ${quote(node.as)}`);
            wheres.push(`${quote(node.as)}.${quote(node.sqlBatch.thisKey.name)} IN (${batchScope.map(_util.maybeQuote).join(',')})`);
          }
        } else {
          (0, _assert2.default)(!parent, `Object type for "${node.fieldName}" table must have a "sqlJoin" or "sqlBatch"`);
          joins.push(`FROM ${node.name} AS ${quote(node.as)}`);
        }

        if (!node.sqlBatch && !node.junctionBatch || !parent) {
          for (let child of node.children) {
            yield _stringifySqlAST(node, child, [...prefix, node.as], context, selections, joins, wheres);
          }
        }

        break;
      case 'column':
        selections.push(`${quote(node.fromOtherTable || parent.as)}.${quote(node.name)} AS ${quote((0, _shared.joinPrefix)(prefix) + node.as)}`);
        break;
      case 'columnDeps':
        for (let name in node.names) {
          selections.push(`${quote(parent.as)}.${quote(name)} AS ${quote((0, _shared.joinPrefix)(prefix) + node.names[name])}`);
        }
        break;
      case 'composite':
        const parentTable = node.fromOtherTable || parent.as;
        const keys = node.name.map(function (key) {
          return `${quote(parentTable)}.${quote(key)}`;
        });

        selections.push(`CONCAT(${keys.join(', ')}) AS ${quote((0, _shared.joinPrefix)(prefix) + node.fieldName)}`);
        break;
      case 'expression':
        const expr = node.sqlExpr(`${quote(parent.as)}`, node.args || {}, context, (0, _shared.quotePrefix)(prefix, '`'));
        selections.push(`${expr} AS ${quote((0, _shared.joinPrefix)(prefix) + node.as)}`);
        break;
      case 'noop':
        return;
      default:
        throw new Error('unexpected/unknown node type reached: ' + (0, _util.inspect)(node));
    }
    return { selections, joins, wheres };
  });

  return function _stringifySqlAST(_x4, _x5, _x6, _x7, _x8, _x9, _x10, _x11) {
    return _ref2.apply(this, arguments);
  };
})();

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _util = require('../util');

var _shared = require('./shared');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

exports.default = (() => {
  var _ref = _asyncToGenerator(function* (topNode, context, batchScope) {
    (0, _util.validateSqlAST)(topNode);

    let { selections, joins, wheres } = yield _stringifySqlAST(null, topNode, [], context, [], [], [], batchScope);

    selections = [...new Set(selections)];

    if (!selections.length) return '';

    let sql = 'SELECT\n  ' + selections.join(',\n  ') + '\n' + joins.join('\n');
    if (wheres.length) {
      sql += '\nWHERE ' + wheres.join(' AND ');
    }
    return sql;
  });

  function stringifySqlAST(_x, _x2, _x3) {
    return _ref.apply(this, arguments);
  }

  return stringifySqlAST;
})();

function quote(str) {
  return '`' + str + '`';
}