'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

let _stringifySqlAST = (() => {
  var _ref2 = _asyncToGenerator(function* (parent, node, prefix, context, selections, joins, wheres, orders, batchScope) {
    switch (node.type) {
      case 'table':
        if (node.where && !node.paginate && (!node.sqlBatch || !parent)) {
          const whereCondition = yield node.where(`"${node.as}"`, node.args || {}, context, (0, _shared.quotePrefix)(prefix));
          if (whereCondition) {
            wheres.push(`${whereCondition}`);
          }
        }

        if (node.sqlJoin) {
          const joinCondition = yield node.sqlJoin(`"${parent.as}"`, `"${node.as}"`, node.args || {}, context);

          if (node.paginate) {
            let whereCondition = yield node.sqlJoin(`"${parent.as}"`, node.name, node.args || {}, context);
            if (node.where) {
              const filterCondition = yield node.where(`${node.name}`, node.args || {}, context, (0, _shared.quotePrefix)(prefix));
              if (filterCondition) {
                whereCondition += ' AND ' + filterCondition;
              }
            }

            if (node.sortKey) {

              const { limit, orderColumns, whereCondition: whereAddendum } = interpretForKeysetPaging(node);
              if (whereAddendum) {
                whereCondition += ' AND ' + whereAddendum;
              }
              const join = `\
LEFT JOIN LATERAL (
  SELECT * FROM ${node.name}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit}
) AS "${node.as}" ON ${joinCondition}`;
              joins.push(join);
              orders.push({
                table: node.as,
                columns: orderColumns
              });
            } else if (node.orderBy) {

              const { limit, offset, orderColumns } = interpretForOffsetPaging(node);
              const join = `\
LEFT JOIN LATERAL (
  SELECT *, count(*) OVER () AS "$total"
  FROM ${node.name}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit} OFFSET ${offset}
) AS "${node.as}" ON ${joinCondition}`;
              joins.push(join);
              orders.push({
                table: node.as,
                columns: orderColumns
              });
            }
          } else {
            joins.push(`LEFT JOIN ${node.name} AS "${node.as}" ON ${joinCondition}`);
          }
        } else if (node.junctionTable && node.junctionBatch) {
          if (parent) {
            selections.push(`"${parent.as}"."${node.junctionBatch.parentKey.name}" AS "${(0, _shared.joinPrefix)(prefix) + node.junctionBatch.parentKey.as}"`);
          } else {
            if (node.paginate) {
              let whereCondition = 'TRUE';
              if (node.where) {
                whereCondition = yield node.where(`${node.name}`, node.args || {}, context, (0, _shared.quotePrefix)(prefix));
              }

              whereCondition += ' AND ' + `${node.junctionTable}."${node.junctionBatch.thisKey.name}" = temp."${node.junctionBatch.parentKey.name}"`;
              if (node.sortKey) {
                const { limit, orderColumns, whereCondition: whereAddendum } = interpretForKeysetPaging(node);
                if (whereAddendum) {
                  whereCondition += ' AND ' + whereAddendum;
                }
                const join = `\
FROM (VALUES ${batchScope.map(function (val) {
                  return `(${val})`;
                })}) temp("${node.junctionBatch.parentKey.name}")
LEFT JOIN LATERAL (
  SELECT * FROM ${node.junctionTable}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit}
) AS "${node.junctionTableAs}" ON "${node.junctionTableAs}"."${node.junctionBatch.thisKey.name}" = temp."${node.junctionBatch.parentKey.name}"`;
                joins.push(join);
                const joinCondition = yield node.junctionBatch.sqlJoin(`"${node.junctionTableAs}"`, node.as, node.args || {}, context);
                joins.push(`LEFT JOIN ${node.name} AS ${node.as} ON ${joinCondition}`);
                orders.push({
                  table: node.junctionTableAs,
                  columns: orderColumns
                });
              } else if (node.orderBy) {

                const { limit, offset, orderColumns } = interpretForOffsetPaging(node);
                const join = `\
FROM (VALUES ${batchScope.map(function (val) {
                  return `(${val})`;
                })}) temp("${node.junctionBatch.parentKey.name}")
LEFT JOIN LATERAL (
  SELECT *, count(*) OVER () AS "$total"
  FROM ${node.junctionTable}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit} OFFSET ${offset}
) AS "${node.junctionTableAs}" ON "${node.junctionTableAs}"."${node.junctionBatch.thisKey.name}" = temp."${node.junctionBatch.parentKey.name}"`;
                joins.push(join);
                const joinCondition = yield node.junctionBatch.sqlJoin(`"${node.junctionTableAs}"`, node.as, node.args || {}, context);
                joins.push(`LEFT JOIN ${node.name} AS ${node.as} ON ${joinCondition}`);
                orders.push({
                  table: node.junctionTableAs,
                  columns: orderColumns
                });
              }
            } else {
              const joinCondition = yield node.junctionBatch.sqlJoin(`"${node.junctionTableAs}"`, `"${node.as}"`, node.args || {}, context);
              joins.push(`FROM ${node.junctionTable} AS "${node.junctionTableAs}"`, `LEFT JOIN ${node.name} AS "${node.as}" ON ${joinCondition}`);
              wheres.push(`"${node.junctionTableAs}"."${node.junctionBatch.thisKey.name}" IN (${batchScope.join(',')})`);
            }
          }
        } else if (node.junctionTable) {
          (0, _assert2.default)(node.sqlJoins, 'Must set "sqlJoins" for a join table.');
          const joinCondition1 = yield node.sqlJoins[0](`"${parent.as}"`, `"${node.junctionTableAs}"`, node.args || {}, context);
          const joinCondition2 = yield node.sqlJoins[1](`"${node.junctionTableAs}"`, `"${node.as}"`, node.args || {}, context);

          if (node.paginate) {
            let whereCondition = yield node.sqlJoins[0](`"${parent.as}"`, node.junctionTable, node.args || {}, context);
            if (node.where) {
              const filterCondition = yield node.where(`${node.name}`, node.args || {}, context, (0, _shared.quotePrefix)(prefix));
              if (filterCondition) {
                whereCondition += ' AND ' + filterCondition;
              }
            }

            if (node.sortKey) {
              const { limit, orderColumns, whereCondition: whereAddendum } = interpretForKeysetPaging(node);
              if (whereAddendum) {
                whereCondition += ' AND ' + whereAddendum;
              }
              const join = `\
LEFT JOIN LATERAL (
  SELECT * FROM ${node.junctionTable}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit}
) AS "${node.junctionTableAs}" ON ${joinCondition1}`;
              joins.push(join);
              orders.push({
                table: node.junctionTableAs,
                columns: orderColumns
              });
            } else if (node.orderBy) {

              const { limit, offset, orderColumns } = interpretForOffsetPaging(node);
              const join = `\
LEFT JOIN LATERAL (
  SELECT *, count(*) OVER () AS "$total"
  FROM ${node.junctionTable}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit} OFFSET ${offset}
) AS "${node.junctionTableAs}" ON ${joinCondition1}`;
              joins.push(join);
              orders.push({
                table: node.junctionTableAs,
                columns: orderColumns
              });
            }
          } else {
            joins.push(`LEFT JOIN ${node.junctionTable} AS "${node.junctionTableAs}" ON ${joinCondition1}`);
          }
          joins.push(`LEFT JOIN ${node.name} AS "${node.as}" ON ${joinCondition2}`);
        } else if (node.sqlBatch) {
          if (parent) {
            selections.push(`"${parent.as}"."${node.sqlBatch.parentKey.name}" AS "${(0, _shared.joinPrefix)(prefix) + node.sqlBatch.parentKey.as}"`);
          } else {
            if (node.paginate) {
              if (node.sortKey) {
                let { limit, orderColumns, whereCondition } = interpretForKeysetPaging(node);
                whereCondition = whereCondition || 'TRUE';
                whereCondition += ' AND ' + `"${node.name}"."${node.sqlBatch.thisKey.name}" = temp."${node.sqlBatch.parentKey.name}"`;
                if (node.where) {
                  const filterCondition = yield node.where(`${node.name}`, node.args || {}, context, []);
                  if (filterCondition) {
                    whereCondition += ' AND ' + filterCondition;
                  }
                }
                const join = `\
FROM (VALUES ${batchScope.map(function (val) {
                  return `(${val})`;
                })}) temp("${node.sqlBatch.parentKey.name}")
JOIN LATERAL (
  SELECT * FROM ${node.name}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit}
) AS "${node.as}" ON "${node.as}"."${node.sqlBatch.thisKey.name}" = temp."${node.sqlBatch.parentKey.name}"`;
                joins.push(join);
                orders.push({
                  table: node.as,
                  columns: orderColumns
                });
              } else if (node.orderBy) {
                const { limit, offset, orderColumns } = interpretForOffsetPaging(node);
                let whereCondition = 'TRUE';
                if (node.where) {
                  const filterCondition = yield node.where(`${node.name}`, node.args || {}, context, []);
                  if (filterCondition) {
                    whereCondition = filterCondition;
                  }
                }
                whereCondition += ' AND ' + `"${node.name}"."${node.sqlBatch.thisKey.name}" = temp."${node.sqlBatch.parentKey.name}"`;
                const join = `\
FROM (VALUES ${batchScope.map(function (val) {
                  return `(${val})`;
                })}) temp("${node.sqlBatch.parentKey.name}")
JOIN LATERAL (
  SELECT *, count(*) OVER () AS "$total"
  FROM ${node.name}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit} OFFSET ${offset}
) AS "${node.as}" ON "${node.as}"."${node.sqlBatch.thisKey.name}" = temp."${node.sqlBatch.parentKey.name}"`;
                joins.push(join);
                orders.push({
                  table: node.as,
                  columns: orderColumns
                });
              }
            } else {
              joins.push(`FROM ${node.name} AS "${node.as}"`);
              wheres.push(`"${node.as}"."${node.sqlBatch.thisKey.name}" IN (${batchScope.join(',')})`);
            }
          }
        } else if (node.paginate) {
          if (node.sortKey) {
            let { limit, orderColumns, whereCondition } = interpretForKeysetPaging(node);
            whereCondition = whereCondition || 'TRUE';
            if (node.where) {
              const filterCondition = yield node.where(`${node.name}`, node.args || {}, context, (0, _shared.quotePrefix)(prefix));
              if (filterCondition) {
                whereCondition += ' AND ' + filterCondition;
              }
            }
            const join = `\
FROM (
  SELECT * FROM ${node.name}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit}
) AS "${node.as}"`;
            joins.push(join);
            orders.push({
              table: node.as,
              columns: orderColumns
            });
          } else if (node.orderBy) {
            const { limit, offset, orderColumns } = interpretForOffsetPaging(node);
            let whereCondition = 'TRUE';
            if (node.where) {
              const filterCondition = yield node.where(`${node.name}`, node.args || {}, context, (0, _shared.quotePrefix)(prefix));
              if (filterCondition) {
                whereCondition = filterCondition;
              }
            }
            const join = `\
FROM (
  SELECT *, count(*) OVER () AS "$total"
  FROM ${node.name}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns)}
  LIMIT ${limit} OFFSET ${offset}
) AS "${node.as}"`;
            joins.push(join);
            orders.push({
              table: node.as,
              columns: orderColumns
            });
          }
        } else {
          (0, _assert2.default)(!parent, `Object type for "${node.fieldName}" table must have a "sqlJoin" or "sqlBatch"`);
          joins.push(`FROM ${node.name} AS "${node.as}"`);
        }

        if (!node.sqlBatch && !node.junctionBatch || !parent) {
          for (let child of node.children) {
            yield _stringifySqlAST(node, child, [...prefix, node.as], context, selections, joins, wheres, orders);
          }
        }

        break;
      case 'column':
        selections.push(`"${node.fromOtherTable || parent.as}"."${node.name}" AS "${(0, _shared.joinPrefix)(prefix) + node.as}"`);
        break;
      case 'columnDeps':
        for (let name in node.names) {
          selections.push(`"${parent.as}"."${name}" AS "${(0, _shared.joinPrefix)(prefix) + node.names[name]}"`);
        }
        break;
      case 'composite':
        const parentTable = node.fromOtherTable || parent.as;
        const keys = node.name.map(function (key) {
          return `"${parentTable}"."${key}"`;
        });
        selections.push(`NULLIF(CONCAT(${keys.join(', ')}), '') AS "${(0, _shared.joinPrefix)(prefix) + node.fieldName}"`);
        break;
      case 'expression':
        const expr = node.sqlExpr(`"${parent.as}"`, node.args || {}, context, (0, _shared.quotePrefix)(prefix));
        selections.push(`${expr} AS "${(0, _shared.joinPrefix)(prefix) + node.as}"`);
        break;
      case 'noop':
        return;
      default:
        throw new Error('unexpected/unknown node type reached: ' + (0, _util.inspect)(node));
    }
    return { selections, joins, wheres, orders };
  });

  return function _stringifySqlAST(_x4, _x5, _x6, _x7, _x8, _x9, _x10, _x11, _x12) {
    return _ref2.apply(this, arguments);
  };
})();

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _graphqlRelay = require('graphql-relay');

var _util = require('../util');

var _shared = require('./shared');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

exports.default = (() => {
  var _ref = _asyncToGenerator(function* (topNode, context, batchScope) {
    (0, _util.validateSqlAST)(topNode);

    let { selections, joins, wheres, orders } = yield _stringifySqlAST(null, topNode, [], context, [], [], [], [], batchScope);

    selections = [...new Set(selections)];

    if (!selections.length) return '';

    let sql = 'SELECT\n  ' + selections.join(',\n  ') + '\n' + joins.join('\n');
    if (wheres.length) {
      sql += '\nWHERE ' + wheres.join(' AND ');
    }
    if (orders.length) {
      sql += '\nORDER BY ' + stringifyOuterOrder(orders);
    }
    return sql;
  });

  function stringifySqlAST(_x, _x2, _x3) {
    return _ref.apply(this, arguments);
  }

  return stringifySqlAST;
})();

function interpretForOffsetPaging(node) {
  if (node.args && node.args.last) {
    throw new Error('Backward pagination not supported with offsets. Consider using keyset pagination instead');
  }
  const orderColumns = {};
  if (typeof node.orderBy === 'object') {
    for (let column in node.orderBy) {
      let direction = node.orderBy[column].toUpperCase();
      if (direction !== 'ASC' && direction !== 'DESC') {
        throw new Error(direction + ' is not a valid sorting direction');
      }
      orderColumns[column] = direction;
    }
  } else if (typeof node.orderBy === 'string') {
    orderColumns[node.orderBy] = 'ASC';
  } else {
    throw new Error('"orderBy" is required for pagination');
  }
  let limit = 'ALL',
      offset = 0;
  if (node.args && node.args.first) {
    limit = parseInt(node.args.first) + 1;
    if (node.args.after) {
      offset = (0, _graphqlRelay.cursorToOffset)(node.args.after) + 1;
    }
  }
  return { limit, offset, orderColumns };
}

function interpretForKeysetPaging(node) {
  const orderColumns = {};
  let descending = node.sortKey.order.toUpperCase() === 'DESC';

  if (node.args && node.args.last) {
    descending = !descending;
  }
  for (let column of (0, _util.wrap)(node.sortKey.key)) {
    orderColumns[column] = descending ? 'DESC' : 'ASC';
  }

  let limit = 'ALL',
      whereCondition = '';
  if (node.args && node.args.first) {
    limit = parseInt(node.args.first) + 1;
    if (node.args.after) {
      const cursorObj = (0, _util.cursorToObj)(node.args.after);
      validateCursor(cursorObj, (0, _util.wrap)(node.sortKey.key));
      whereCondition = sortKeyToWhereCondition(cursorObj, descending);
    }
    if (node.args.before) {
      throw new Error('Using "before" with "first" is nonsensical.');
    }
  } else if (node.args && node.args.last) {
    limit = parseInt(node.args.last) + 1;
    if (node.args.before) {
      const cursorObj = (0, _util.cursorToObj)(node.args.before);
      validateCursor(cursorObj, (0, _util.wrap)(node.sortKey.key));
      whereCondition = sortKeyToWhereCondition(cursorObj, descending);
    }
    if (node.args.after) {
      throw new Error('Using "after" with "last" is nonsensical.');
    }
  }

  return { limit, orderColumns, whereCondition };
}

function sortKeyToWhereCondition(keyObj, descending) {
  const sortColumns = [];
  const sortValues = [];
  for (let key in keyObj) {
    sortColumns.push(`"${key}"`);
    sortValues.push((0, _util.maybeQuote)(keyObj[key]));
  }
  const operator = descending ? '<' : '>';
  return `(${sortColumns.join(', ')}) ${operator} (${sortValues.join(', ')})`;
}

function orderColumnsToString(orderColumns) {
  const conditions = [];
  for (let column in orderColumns) {
    conditions.push(`"${column}" ${orderColumns[column]}`);
  }
  return conditions.join(', ');
}

function stringifyOuterOrder(orders) {
  const conditions = [];
  for (let condition of orders) {
    for (let column in condition.columns) {
      const direction = condition.columns[column];
      conditions.push(`"${condition.table}"."${column}" ${direction}`);
    }
  }
  return conditions.join(', ');
}

function validateCursor(cursorObj, expectedKeys) {
  const actualKeys = Object.keys(cursorObj);
  const expectedKeySet = new Set(expectedKeys);
  const actualKeySet = new Set(actualKeys);
  for (let key of actualKeys) {
    if (!expectedKeySet.has(key)) {
      throw new Error(`Invalid cursor. The column "${key}" is not in the sort key.`);
    }
  }
  for (let key of expectedKeys) {
    if (!actualKeySet.has(key)) {
      throw new Error(`Invalid cursor. The column "${key}" is not in the cursor.`);
    }
  }
}